(function($) {

    RAVEN.namespace("RAVEN.PreCondition");

    /**
     * This is the main view for the Test Runs table on a Test Execution issue
     */
    RAVEN.PreCondition.TestsView = Backbone.View.extend({
        el: '#raven-precondition-tests',

        /**
         * DataTable Instance
         */
        dataTable: null,

        /**
         * The PreCondition issue key
         */
        precondKey: null,

        /**
         * Permissions
         */
        canEdit: false,

        /**
         * The Bulk view
         */
        bulkEditView: null,

        events: {
            'click #raven-precondition-tests #raven-add-tests': '_onAddTestsClick',
            'click span.remove-command': '_onTestDeleteClick'
        },

        /**
         * Unscoped Events - these events must be registered elsewhere as they will not be
         * fired by backbone because they are not within the view element. Refer to "_delegateUnscopedEvents".
         */
        unscopedEvents: {
            'click.raven.precond .bulk-remove-command': '_onBulkTestDeleteClick'
        },

        initialize: function() {
            AJS.messages.setup();

            this._clearMessage();

            this._delegateUnscopedEvents();

            this.precondKey = $('#raven-precondition-tests-issue-key').val();

            this.canEdit = $('#can-edit-precondition-test').val() === 'true';

            this._setEditable(this.canEdit);

            if (this.$el.closest(document.documentElement).length > 0) {
                this.render();
            }
        },

        render: function() {
            this._clearMessage();
            
            // create the dataTable
            this._createTestsDataTable();

            // render bulk mode button
            if (this.canEdit === true && !this.bulkEditView) {
                this._renderBulkModeButton();
            }

            this.$el.show();

            return this;
        },

        destroy: function() {
            this._cleanView();

            // COMPLETELY UNBIND THE VIEW
            this.undelegateEvents();

            this._undelegateUnscopedEvents();

            this.$el.removeData().unbind();

            this.unbind();

            this.remove();
        },

        _cleanView: function() {
            if (this.dataTable) {
                this.dataTable.fnDestroy(true);
                this.dataTable = null;
            }

            this.bulkEditView = null;
        },

        message: function (title, message, type) {
            type = (type === undefined || type == "") ? "error" : type;

            if (!message && title) {
                message = title;
                title = null;
            }

            if (type === "error") {
                this._clearMessage();
                AJS.messages.error('#raven-tests-message-bar', {
                    title: title,
                    body: message
                });
            }

            if (type === "success") {
                AJS.messages.success('#raven-tests-message-bar', {
                    title: title,
                    body: message
                });
            }

            if (type === "info") {
                AJS.messages.info('#raven-tests-message-bar', {
                    title: title,
                    body: message
                });
            }
        },

        _clearMessage: function() {
            $('#raven-tests-message-bar').empty();
        },

        _setEditable: function(editable) {
            if (editable == false) {
                $('#raven-add-tests').hide();
            } else {
                $('#raven-add-tests').show();
            }
        },

        _createTestsDataTable: function() {
            var self = this;

            var path = contextPath + '/rest/raven/1.0/precondition/' + this.precondKey + "/tests";

            if(this.dataTable) {
                this.dataTable.fnDestroy();
                //$('#exec-entries-table').empty();
            }

            this.dataTable = this.$el.find('.tests-table').xrayDataTable( {
                "bProcessing": false,
                "oLanguage": {
                    "oPaginate": {
                        "sFirst": AJS.I18n.getText('common.label.first'),
                        "sLast": AJS.I18n.getText('common.label.last'),
                        "sNext": AJS.I18n.getText('common.label.next'),
                        "sPrevious": AJS.I18n.getText('common.label.previous')
                    },
                    "sEmptyTable": AJS.I18n.getText('view-issue-precondition-tests-section.emptytable'),
                    "sInfoEmpty": AJS.I18n.getText('view-issue-precondition-tests-section.infoempty'),
                    "sInfo": AJS.I18n.getText('common.table.label.showing-records'),
                    "sZeroRecords": AJS.I18n.getText('common.table.label.no-records-found'),
                    "sInfoFiltered": AJS.I18n.getText('common.table.label.table-filtered-from'),
                    "sSearch":  AJS.I18n.getText('common.table.label.search'),
                    "sLengthMenu": AJS.I18n.getText('common.table.label.show-menu-entries')
                },
                "bSort": true,
                "aaSorting": [],
                "sAjaxSource": path,
                "sPaginationType": "full_numbers",
                "bServerSide" : true,
                "bRetrieve" : true,
                "bAutoWidth": false,
                "sAjaxDataProp": "entries",
                "aoColumnDefs": this._createDataTableColumnDefs(),
                "fnServerData": function ( sSource, aoData, fnCallback, oSettings ) {
                    self.$el.block();
                    oSettings.jqXHR = $.ajax( {
                        "dataType": 'json',
                        "type": "GET",
                        "url": sSource,
                        "data": aoData,
                        "error": function(qXHR, textStatus, errorThrown) {
                            if (errorThrown && errorThrown.trim() != '') {
                                errorThrown = errorThrown + ': ' + qXHR.responseText;
                            } else {
                                errorThrown = qXHR.responseText;
                            }
                            self.message(AJS.I18n.getText('view-issue-precondition-tests-section.processingerror'), errorThrown)
                        },
                        "complete": function() {
                            self.$el.unblock();
                        },
                        "success": fnCallback
                    });

                    if (self.bulkEditView) {
                        self.bulkEditView.setCheckBoxRowsSelectionValue(false);
                    }
                },
                "fnRowCallback": function( nRow, aData, iDisplayIndex, iDisplayIndexFull ) {
                    nRow.id = aData["key"];
                    return nRow;
                },
                "fnDrawCallback": function( oSettings ) {
                    self._toogleTestsTable();
                    if (self.bulkEditView) {
                        self.bulkEditView.showHideBulkActions();
                    }
                }
            });
        },

        /**
         * Creates the DataTable column def descriptors
         */
        _createDataTableColumnDefs: function() {
            var self = this;

            var bulkToogleColumnTitle = null;
            var bulkActionColumnTitle = null;
            if (self.canEdit === true) {
                bulkToogleColumnTitle = RAVEN.DataTable.BulkEdit.renderBulkSelectAllInput();

                var bulkActions = '<a href="javascript:void(0);" data-key="bulkKey" class="bulk-remove-command active aui-dropdown2-active" title="' + AJS.I18n.getText('common.action.remove') + '">'+
                                    '<span class="icon icon-delete menu-icon"></span>' + AJS.I18n.getText('common.action.remove') + '</a>';
                bulkActionColumnTitle = RAVEN.DataTable.BulkEdit.renderBulkActionTrigger({id: "tests-bulk-action",actions: bulkActions});
            }

            return [
                {
                    "mData": null,
                    "sTitle": bulkToogleColumnTitle,
                    "mDataProp": null,
                    "sWidth": "20px",
                    "sDefaultContent": "<input type='checkbox' ></input>",
                    "bSortable": false,
                    "bSearchable": false,
                    "mRender": function ( data, type, row ) {
                        if (self.canEdit === true) {
                            return RAVEN.DataTable.BulkEdit.renderBulkSelectInput({id: row["key"]});
                        } else {
                            return '';
                        }
                    },
                    "aTargets": [ 0 ]
                },
                {
                    "mData": "key",
                    "bSortable": true,
                    "mRender": function ( data, type, row ) {
                        return '<a href=' + contextPath + '/browse/' + row["key"] + ' class="link-title">' + data + '</a>';
                    },
                    "aTargets": [ 1 ]
                },
                {
                    "mData": "summary",
                    "bSortable": true,
                    "mRender": function ( data, type, row ) {
                        return '<a href=' + contextPath + '/browse/' + row["key"] + ' class="link-title">' + data + '</a>';
                    },
                    "aTargets": [ 2 ]
                },
                {
                    "mData": null,
                    "sTitle": bulkActionColumnTitle,
                    "bSortable": false,
                    "bSearchable": false,
                    "mRender": function ( data, type, row ) {
                        if (self.canEdit == false) {
                            return '<span id="test_row_'+new Date().getTime()+'"></span>';
                        } else {
                            return '<span id="test_row_'+new Date().getTime()+'" class="icon icon-delete remove-command"></span>';
                        }
                    },
                    "aTargets": [ 3 ]
                }
            ];
        },

        _toogleTestsTable: function() {
            if (this.dataTable && this.dataTable.fnRecordsTotal() > 0) {
                this.$el.find('.notests').hide();
                this.$el.find('.xrayDataTables_wrapper').show();
            } else {
                this.$el.find('.xrayDataTables_wrapper').hide();
                this.$el.find('.notests').show();
            }
        },

        _renderBulkModeButton: function() {
            var $wrapper = this.$el.find('.xrayDataTables_wrapper');
            // if filter already exists then return
            if ($wrapper.find('#bulk-edit-action').length > 0) {
                return;
            }

            this.bulkEditView = new RAVEN.DataTable.BulkEditView({
                el: $wrapper
            });
        },

        _addTests: function(testKeys) {
            var self = this;

            this.$el.block();

            var value = JSON.stringify(testKeys);
            $.ajax({
                type: "POST",
                url: contextPath + '/rest/raven/1.0/precondition/' + this.precondKey + '/tests',
                processData: true,
                async: true,
                data: value,
                contentType: "application/json",
                error: function(qXHR, textStatus, errorThrown) {
                    if (errorThrown && errorThrown.trim() != '') {
                        errorThrown = errorThrown + ': ' + qXHR.responseText;
                    } else {
                        errorThrown = qXHR.responseText;
                    }

                    self.message(AJS.I18n.getText('view-issue-precondition-tests-section.add.error', self.precondKey), errorThrown);
                },
                success: function(data, textStatus, jqXHR) {
                    if (data.errors && data.errors.length > 0) {
                        var errorMsg = '';
                        $.each($(data.errors), function(i, msg) {
                            errorMsg = errorMsg + msg;
                            if (i < data.errors.length) {
                                errorMsg = errorMsg + '<br/>';
                            }
                        });
                        self.message(AJS.I18n.getText('view-issue-precondition-tests-section.add.controlled.error', self.precondKey), errorMsg);
                    }

                    if (data.tests && data.tests.length > 0) {
                        self.dataTable.fnAddData(data.tests);
                        // JIRA.trigger(JIRA.Events.REFRESH_ISSUE_PAGE, [JIRA.Issue.getIssueId()]);
                    }
                },
                complete: function() {
                    self.$el.unblock();
                    self._updateHistory();
                    self._updateDates();
                }
            });
        },

        _confirmationDeleteTest: function(testKeys, yesFunction, bulk) {
            var self = this;

            var dialog = new JIRA.FormDialog({
                id: 'delete-test-dialog',
                content: function (callback) {
                    var title = null;
                    var message = null;
                    if (bulk === true) {
                        title = AJS.I18n.getText('view-issue-precondition-tests-section.bulk.delete.confirm.title');
                        message = AJS.I18n.getText('view-issue-precondition-tests-section.bulk.delete.confirm.message', self.precondKey);
                    } else {
                        title = AJS.I18n.getText('view-issue-precondition-tests-section.delete.confirm.title', testKeys[0]);
                        message = AJS.I18n.getText('view-issue-precondition-tests-section.delete.confirm.message', self.precondKey);
                    }
                    var severity = "info";
                    var innerHtml = RAVEN.Dialog.deleteConfirmationDialog({title:title, message:message, severity:severity});
                    callback(innerHtml);
                },
                submitHandler: function (e) {
                    e.preventDefault();
                    var proxyFunc = $.proxy(yesFunction, self);
                    proxyFunc(testKeys);
                    dialog.hide();
                },
                autoClose : true
            });
            dialog.show();
        },

        _deleteTest: function(testKeys) {
            var self = this;

            this.$el.block();

            //if we have records to delete
            if(testKeys && testKeys.length > 0) {
                //Build query param
                var testKeysParam = this._buildQueryParamsFromSelectedRows(testKeys);

                $.ajax({
                    type: "DELETE",
                    url: contextPath + '/rest/raven/1.0/precondition/' + self.precondKey + "/tests/delete/?" + testKeysParam,
                    processData: true,
                    async: true,
                    contentType: "application/json",
                    error: function(qXHR, textStatus, errorThrown) {
                        if (errorThrown && errorThrown.trim() != '') {
                            errorThrown = errorThrown + ': ' + qXHR.responseText;
                        } else {
                            errorThrown = qXHR.responseText;
                        }
                        self.message(AJS.I18n.getText('view-issue-precondition-tests-section.delete.error'), errorThrown);
                    },
                    success: function(data, textStatus, jqXHR) {
                        self._deleteTestsCallback(testKeys, data);
                    },
                    complete: function() {
                        $('#raven-exec-tests').unblock();
                        self._updateHistory();
                        self._updateDates();
                    }
                });
            }
        },

        /**
         * Build query param
         */
        _buildQueryParamsFromSelectedRows: function(selectedIds) {
            var testKeys = "";
            var i;
            for (i = 0; i < selectedIds.length; ++i) {
                if(testKeys.length > 0){
                    testKeys+="&";
                }
                testKeys += "testKeys=" + selectedIds[i];
            }

            return testKeys;
        },

        _deleteTestsCallback: function(testKeys, data) {
            var self = this;
            if (testKeys && testKeys.length > 0) {
                var allData = this.dataTable.fnGetData();
                if (allData.length >= 1) {
                    _.each(_.initial(testKeys), function(testKey) {
                        self.dataTable.fnDeleteRow(testKey, null, false);
                    });
                    this.dataTable.fnDeleteRow(_.last(testKeys));
                } else {
                    // don't refresh the table if the last element is removed because it will request more data from server
                    this.dataTable.fnDeleteRow(_.last(testKeys), null, false);
                }
            }
        },

        _updateHistory: function() {
            if ($('#changehistory-tabpanel').attr('class') == 'active') {
                var alt_token = $('input[name="atl_token"]').val();
                $.ajax({
                    type : "GET",
                    url : contextPath + '/browse/' + this.precondKey +"?page=com.atlassian.jira.plugin.system.issuetabpanels:changehistory-tabpanel&_pjax=true&_="+alt_token,
                    contentType : "text/plain",
                    success : function (data, status) {
                        var tab = $(data).find('#activitymodule div.mod-content #issue_actions_container');
                        var jiraTab = $('#activitymodule div.mod-content').find('#issue_actions_container');
                        $(jiraTab).html(tab);
                    }
                });
            }
        },

        _updateDates: function () {
            var updatedDate = null;

            if(AJS.$('#datesmodule #updated-date').length === 1) {
                updatedDate = AJS.$('#datesmodule #updated-date time.livestamp');
            } else if(AJS.$('#datesmodule #updated-val').length === 1) {
                updatedDate = AJS.$('#datesmodule #updated-val time.livestamp');
            } else {
                return null;
            }

            $.ajax({
                type : "GET",
                url : AJS.contextPath() + '/rest/raven/1.0/date/' + JIRA.Issue.getIssueKey() + "/updated",
                contentType : "application/json",
                success : function (data) {
                    var title = data.title;
                    var dateTimeIso8601 = data.dateTimeIso8601;
                    updatedDate.attr("datetime", dateTimeIso8601);
                    $.each(AJS.$('#datesmodule dd.date'), function() {
                        if($(this).find('#updated-date time.livestamp').length > 0 || $(this).find('#updated-val time.livestamp').length > 0) {
                            $(this).attr("title", title);
                        }
                    });
                    updatedDate.livestamp();
                }
            });
        },

        /****************************************************************
         *                        EVENT HANDLERS
         ****************************************************************/

        _delegateUnscopedEvents: function() {
            var self = this;
            $(document).off('.raven.precond');

            $.each(this.unscopedEvents, function(key, val) {
                var eventName = key.substring(0, key.indexOf(' '));
                var eventTrigger = key.substring(key.indexOf(' ') + 1);

                $(document).on(eventName, eventTrigger, $.proxy(self[val], self));
            });
        },

        _undelegateUnscopedEvents: function() {
            $(document).off('.raven.precond');
        },

        /**
         * Delete test click handler
         */
        _onTestDeleteClick: function(e) {
            e.stopPropagation();
            e.preventDefault();

            this._clearMessage();

            var $row = $(e.target).closest('tr');
            var testData = this.dataTable.fnGetData($row[0]);

            this._confirmationDeleteTest([ testData.key ], this._deleteTest);
        },

        /**
         * Delete Bulk test click handler
         */
        _onBulkTestDeleteClick: function(e) {
            e.stopPropagation();
            e.preventDefault();

            this._clearMessage();

            var rowIds = this.bulkEditView.getSelectedRowIds();

            if(rowIds && rowIds.length > 0){
                this._confirmationDeleteTest(rowIds, this._deleteTest, true);
            }
        },

        /**
         * Add tests button click handler
         */
        _onAddTestsClick: function() {
            var self = this;

            this._clearMessage();

            var labels = {};
            labels.label = AJS.I18n.getText('dialogs.issueselector.test.label');
            labels.title = AJS.I18n.getText('dialogs.issueselector.test.title', this.precondKey);
            labels.description = AJS.I18n.getText('dialogs.issueselector.test.description');
            labels.enter = AJS.I18n.getText('dialogs.issueselector.test.enter');

            var issueTypeId = $('#test-issuetype-id').val();

            new RAVEN.Dialogs.createIssueSelectorAdvancedDialog('TEST', 'key not in preConditionTests(' + this.precondKey + ')', issueTypeId, this.precondKey, labels, function(issues) {
                if (issues && issues.length > 0) {
                    self._addTests(issues);
                }
            }, null);
        }
    });
})(AJS.$);
