(function ($) {

    /**
     * A table who's entries/rows are can be retrieved, added and updated via rest (CRUD).
     * It uses backbone.js to sync the tables state back to the server and vice versa, avoiding page refreshes.
     *
     * @class RestfulTable
     */
    window.RAVEN = window.RAVEN || {};
    
    RAVEN.RestfulTable = Backbone.View.extend({
        /**
         * @constructor
         * @param {Object} options
         * ... {String} id - The id for the table. This id will be used to fire events specific to this instance.
         * ... {boolean} allowEdit - Is the table editable. If true, clicking row will switch it to edit state.
         * ... {boolean} allowClone - Are the table rows cloneable? If true, a clone button is added to options.
         * ... {boolean} allowCreate - Can new entries be added to the table.
         * ... {boolean} allowReorder - Can we drag rows to reorder them.
         * ... {String} noEntriesMsg - Message that will be displayed under the table header if it is empty.
         * ... {Array} entries - initial data set to be rendered. Each item in the array will be used to create a new instance of options.model.
         * ... {RAVEN.RestfulTable.EntryModel} model - backbone model representing a row.
         * ... {Object} views
         * ... ... {RAVEN.RestfulTable.EditRow} editRow - Backbone view that renders the edit & create row. Your view MUST extend RAVEN.RestfulTable.EditRow.
         * ... ... {RAVEN.RestfulTable.Row} row - Backbone view that renders the readonly row. Your view MUST extend RAVEN.RestfulTable.Row.
         */
        initialize: function (options) {

            var instance = this;

            // combine default and user options
            instance.options = $.extend(true, instance._getDefaultOptions(options), options);

            // Prefix events for this instance with this id.
            instance.id = this.options.id;

            // faster lookup
            instance._event = RAVEN.RestfulTable.Events;
            instance.classNames = RAVEN.RestfulTable.ClassNames;
            instance.dataKeys = RAVEN.RestfulTable.DataKeys;
            instance.sameOrder = (this.options.remainOrder != undefined && this.options.remainOrder != "") ? this.options.remainOrder : false;
            
            // shortcuts to popular elements
            this.$table = $(options.el)
                    .addClass(this.classNames.RESTFUL_TABLE)
                    .addClass(this.classNames.ALLOW_HOVER)
                    .addClass("aui")
                    .addClass(instance.classNames.LOADING);

            this.$table.wrapAll("<form class='aui' action='#' />");

            this.$thead = $("<thead/>");
            this.$theadRow = $("<tr />").appendTo(this.$thead);
            this.$tbody = $("<tbody/>");

            if (!this.$table.length) {
                throw new Error("RAVEN.RestfulTable: Init failed! The table you have specified [" + this.$table.selector + "] cannot be found.")
            }

            if (!this.options.columns) {
                throw new Error("RAVEN.RestfulTable: Init failed! You haven't provided any columns to render.")
            }

            // Let user know the table is loading
            this.showGlobalLoading();

            $.each(this.options.columns, function (i, column) {
                var header = $.isFunction(column.header) ? column.header() : column.header;
                if (typeof header === "undefined") {
                    console.warn("You have not specified [header] for column [" + column.id + "]. Using id for now...");
                    header = column.id;
                }

                instance.$theadRow.append("<th>" + header + "</th>");
            });

            // columns for submit buttons and loading indicator used when editing
            instance.$theadRow.append("<th></th><th></th>");

            // create a new Backbone collection to represent rows (http://documentcloud.github.com/backbone/#Collection)
            this._models = this._createCollection();

            // shortcut to the class we use to create rows
            this._rowClass = this.options.views.row;

            this.editRows = []; // keep track of rows that are being edited concurrently

            this.$table.closest("form").submit(function (e) {
                if (instance.focusedRow) {
                    // Delegates saving of row. See RAVEN.RestfulTable.EditRow.submit
                    instance.focusedRow.trigger(instance._event.SAVE);
                }
                e.preventDefault();
            });

            if (this.options.allowReorder) {

                // Add allowance for another cell to the thead
                this.$theadRow.prepend("<th />");

                // Allow drag and drop reordering of rows
                this.$tbody.sortable({
                    handle: "." +this.classNames.DRAG_HANDLE,
                    placeholder: "drag-placeholder",
                    helper: function(e, elt) {
                        var helper =  elt.clone(true).addClass(instance.classNames.MOVEABLE);
                        helper.children().each(function (i) {
                            $(this).width(elt.children().eq(i).width());
                        });
                        return helper;
                    },
                    start: function (event, ui) {
                        var $this = ui.placeholder.find("td");
                        // Make sure that when we start dragging widths do not change
                        ui.item
                                .addClass(instance.classNames.MOVEABLE)
                                .children().each(function (i) {
                                    $(this).width($this.eq(i).width());
                                });

                        var row = instance.getRowFromElement(ui.item[0]);

                        // Add a <td> to the placeholder <tr> to inherit CSS styles.
                        ui.placeholder
                                .html('<td colspan="' + instance.getColumnCount() + '">&nbsp;</td>')
                                .css("visibility", "visible")
                                .css('height', row.$el.height());

                        // Stop hover effects etc from occuring as we move the mouse (while dragging) over other rows
                        instance.getRowFromElement(ui.item[0]).trigger(instance._event.MODAL);
                    },
                    stop: function (event, ui) {
                        if (jQuery(ui.item[0]).is(":visible")) {
                            ui.item
                                    .removeClass(instance.classNames.MOVEABLE)
                                    .children().attr("style", "");

                            ui.placeholder.removeClass(instance.classNames.ROW);

                            // Return table to a normal state
                            instance.getRowFromElement(ui.item[0]).trigger(instance._event.MODELESS);
                        }
                    },
                    update: function (event, ui) {
                        var nextModel,
                                nextRow,
                                data = {},
                                row = instance.getRowFromElement(ui.item[0]);

                        if (row) {
                            if (instance.options.reverseOrder) {
                                // Everything is backwards here because on the client we are in reverse order.
                                nextRow = ui.item.next();
                                if (!nextRow.length) {
                                    data.position = "First";
                                } else {
                                    nextModel = instance.getRowFromElement(nextRow).model;
                                    data.after = nextModel.url();
                                }
                            } else {
                                nextRow = ui.item.prev();
                                if (!nextRow.length) {
                                    data.position = "First";
                                } else {
                                    nextModel = instance.getRowFromElement(nextRow).model;
                                    data.after = nextModel.url();
                                }
                            }

                            $.ajax({
                                url: row.model.url() + "/move",
                                type: "POST",
                                dataType: "json",
                                contentType: "application/json",
                                data: JSON.stringify(data),
                                complete: function () {
                                    // hides loading indicator (spinner)
                                    row.hideLoading();
                                },
                                success: function (xhr) {
                                    if (instance.sameOrder) {
                                        instance._reindexRows();
                                    }

                                    AJS.triggerEvtForInst(instance._event.REORDER_SUCCESS, instance, [xhr]);
                                },
                                error: function (xhr) {
                                    var responseData = $.parseJSON(xhr.responseText || xhr.data);
                                    AJS.triggerEvtForInst(instance._event.SERVER_ERROR, instance, [responseData, xhr]);
                                }
                            });

                            // shows loading indicator (spinner)
                            row.showLoading();
                        }
                    },
                    axis: "y",
                    delay: 0,
                    cursor: "move",
                    scroll: true,
                    zIndex: 8000
                });

                // Prevent text selection while reordering.
                this.$tbody.bind("selectstart mousedown", function (event) {
                    return !$(event.target).is("." + instance.classNames.DRAG_HANDLE);
                });
            }


            if (this.options.allowCreate !== false) {

                // Create row responsible for adding new entries ...
                this._createRow = new this.options.views.editRow({
                    columns: this.options.columns,
                    isCreateRow: true,
                    model: this.options.model.extend({
                        url: function () {
                            return instance.options.resources.self;
                        }
                    }),
                    cancelAccessKey: this.options.cancelAccessKey,
                    submitAccessKey: this.options.submitAccessKey,
                    allowReorder: this.options.allowReorder
                })
                        .bind(this._event.CREATED, function (values) {
                            if (instance.options.createPosition === "bottom") {
                                instance.addRow(values);
                            } else {
                                instance.addRow(values, 0);
                            }
                        })
                        .bind(this._event.VALIDATION_ERROR, function () {
                            this.trigger(instance._event.FOCUS);
                        })
                        .render({
                            errors: {},
                            values: {}
                        });

                // ... and appends it as the first row
                this.$create = $('<tbody class="' + this.classNames.CREATE + '" />')
                        .append(this._createRow.el);

                // Manage which row has focus
                this._applyFocusCoordinator(this._createRow);

                // focus create row
                this._createRow.trigger(this._event.FOCUS);
            }

            // when a model is removed from the collection, remove it from the viewport also
            this._models.bind("remove", function (model) {
                $.each(instance.getRows(), function (i, row) {
                    if (row.model === model) {
                        if (row.hasFocus() && instance._createRow) {
                            instance._createRow.trigger(instance._event.FOCUS);
                        }
                        instance.removeRow(row);
                    }
                });
            });

            if ($.isFunction(this.options.resources.all)) {
                this.options.resources.all(function (entries) {
                    instance.populate(entries);
                });
            } else {
                $.get(this.options.resources.all, function (entries) {
                    instance.populate(entries);
                });
            }
        },

        _createCollection: function() {
            var instance = this;

            // create a new Backbone collection to represent rows (http://documentcloud.github.com/backbone/#Collection)
            var rowsAwareCollection = this.options.Collection.extend({
                // Force the collection to re-sort itself. You don't need to call this under normal
                // circumstances, as the set will maintain sort order as each item is added.
                sort:function (options) {
                    options || (options = {});
                    if (!this.comparator) {
                        throw new Error('Cannot sort a set without a comparator');
                    }
                    this.tableRows = instance.getRows();
                    this.models = this.sortBy(this.comparator);
                    this.tableRows = undefined;
                    if (!options.silent) {
                        this.trigger('refresh', this, options);
                    }
                    return this;
                },
                remove:function (models, options) {
                    this.tableRows = instance.getRows();
                    Backbone.Collection.prototype.remove.apply(this, arguments);
                    this.tableRows = undefined;
                    return this;
                }
            });

            return new rowsAwareCollection([], {
                comparator:function (row) {
                    // sort models in collection based on dom ordering
                    var index;
                    $.each(this.tableRows !== undefined ? this.tableRows : instance.getRows(), function (i) {
                        if (this.model.id === row.id) {
                            index = i;
                            return false;
                        }
                    });
                    return index;
                }
            });
        },

        /**
         * Refreshes table with entries
         *
         * @param entries
         */
        populate: function (entries) {
            var rowViews = [];

            if (this.options.reverseOrder) {
                entries.reverse();
            }

            this.hideGlobalLoading();
            if (entries && entries.length) {
                // Empty the models collection
                this._models.reset([], { silent: true });
                // Add all the entries to collection and render them
                rowViews = this.renderRows(entries);
                // show message to user if we have no entries
                if (this.isEmpty()) {
                    this.showNoEntriesMsg();
                }
            } else {
                this.showNoEntriesMsg();
            }

            // Ok, lets let everyone know that we are done...
            this.$table
                    .append(this.$thead);

            if (this.options.createPosition === "bottom") {
                this.$table.append(this.$tbody)
                        .append(this.$create);
            } else {
                this.$table
                        .append(this.$create)
                        .append(this.$tbody);
            }

            this.$table.removeClass(this.classNames.LOADING)
                    .trigger(this._event.INITIALIZED, [this]);

            AJS.triggerEvtForInst(this._event.INITIALIZED, this, [this]);

            // RAVEN: fire append to body event on rows
            var instance = this;
            $.each(rowViews, function() {
                // fire ROW_APPENDED_TO_DOM on the row itself
                this.trigger(instance._event.ROW_APPENDED_TO_DOM);
            });


            if (this.options.autoFocus) {
                this.$table.find(":input:text:first").focus(); // set focus to first field
            }
        },

        /**
         * Shows loading indicator and text
         *
         * @return {RAVEN.RestfulTable}
         */
        showGlobalLoading: function () {

            if (!this.$loading) {
                this.$loading =  $('<div class="raven-restfultable-init"><span class="raven-restfultable-throbber">' +
                        '</span><span class="raven-restfultable-loading">' + this.options.loadingMsg + '</span></div>');
            }
            if (!this.$loading.is(":visible")) {
                this.$loading.insertAfter(this.$table);
            }

            return this
        },

        /**
         * Hides loading indicator and text
         * @return {RAVEN.RestfulTable}
         */
        hideGlobalLoading: function () {
            if (this.$loading) {
                this.$loading.remove();
            }
            return this;
        },


        /**
         * Adds row to collection and renders it
         *
         * @param {Object} values
         * @param {number} index
         * @return {RAVEN.RestfulTable}
         */
        addRow: function (values, index) {

            var view,
                    model;

            if (!values.id) {
                throw new Error("RAVEN.RestfulTable.addRow: to add a row values object must contain an id. "
                        + "Maybe you are not returning it from your restend point?"
                        + "Recieved:" + JSON.stringify(values));
            }

            model = new this.options.model(values);
            this._models.add(model);

            view = this._renderRow(model, index);

            
            this.removeNoEntriesMsg();

            // Let everyone know we added a row
            AJS.triggerEvtForInst(this._event.ROW_ADDED, this, [view, this]);
            return this;
        },

        /**
         * Provided a view, removes it from display and backbone collection
         *
         * @param {RAVEN.RestfulTable.Row}
         */
        removeRow: function (row) {
            this._models.remove(row.model);
            row.remove();

            if (this.isEmpty()) {
                this.showNoEntriesMsg();
            }

            if (this.sameOrder) {
                this._reindexRows();
            }
            // Let everyone know we removed a row
            AJS.triggerEvtForInst(this._event.ROW_REMOVED, this, [row, this]);
        },

        /**
         * Removes row from collection and renders it in a new position
         *
         * @param {RAVEN.RestfulTable.Row}
         * @param {number} index
         */
        moveRow: function(row, index) {
            var position = parseInt(index);

            row.remove();
            this._renderRow(row.model, position - 1); //rank starts in 1, so we have to decrease 1 in order to match the table position

            if(this.sameOrder) {
                this._reindexRows();
            }
        },

        /**
         * Is there any entries in the table
         *
         * @return {Boolean}
         */
        isEmpty: function () {
            return this._models.length === 0;
        },

        /**
         * Gets all models
         *
         * @return {Backbone.Collection}
         */
        getModels: function () {
            return this._models;
        },

        /**
         * Gets table body
         *
         * @return {jQuery}
         */
        getTable: function () {
            return this.$table;
        },

        /**
         * Gets table body
         *
         * @return {jQuery}
         */
        getTableBody: function () {
            return this.$tbody;
        },

        /**
         * Gets create Row
         *
         * @return {B
         */
        getCreateRow: function () {
            return this._createRow;
        },

        /**
         * Gets the number of table colums
         *
         * @return {Number}
         */
        getColumnCount: function () {
            return this.options.columns.length + 2; // plus 2 accounts for the columns allocated to submit buttons and loading indicator
        },

        /**
         * Get the RAVEN.RestfulTable.Row that corresponds to the given <tr> element.
         *
         * @param {HTMLElement} tr
         * @return {?RAVEN.RestfulTable.Row}
         */
        getRowFromElement: function (tr) {
            return $(tr).data(this.dataKeys.ROW_VIEW);
        },

        /**
         * Shows message {options.noEntriesMsg} to the user if there are no entries
         *
         * @return {RAVEN.RestfulTable}
         */
        showNoEntriesMsg: function () {

            if (this.$noEntries) {
                this.$noEntries.remove();
            }

            this.$noEntries = $("<tr>")
                    .addClass(this.classNames.NO_ENTRIES)
                    .append($("<td>")
                    .attr("colspan", this.getColumnCount())
                    .text(this.options.noEntriesMsg)
            )
                    .appendTo(this.$tbody);

            return this;
        },

        /**
         * Removes message {options.noEntriesMsg} to the user if there ARE entries
         *
         * @return {RAVEN.RestfulTable}
         */
        removeNoEntriesMsg: function () {
            if (this.$noEntries && this._models.length > 0) {
                this.$noEntries.remove();
            }
            return this;
        },

        /**
         * Gets the RAVEN.RestfulTable.Row from their associated <tr> elements
         *
         * @return {Array<RAVEN.RestfulTable.Row>}
         */
        getRows: function () {

            var instance = this,
                    views = [];

            this.$tbody.find("." + this.classNames.READ_ONLY).each(function () {

                var $row = $(this),
                        view = $row.data(instance.dataKeys.ROW_VIEW);

                if (view) {
                    views.push(view);
                }
            });

            return views;
        },

        /**
         * Appends entry to end or specified index of table
         *
         * @param {RAVEN.RestfulTable.EntryModel} model
         * @param index
         * @return {jQuery}
         */
        _renderRow: function (model, index) {
            var instance = this,
                    $rows = this.$tbody.find("." + this.classNames.READ_ONLY),
                    $row,
                    view;

            var cloneCallback = function (entry){
                instance.addRow(entry,undefined);
                
            };

            view = new this._rowClass({
                model: model,
                columns: this.options.columns,
                allowEdit: this.options.allowEdit,
                allowClone: this.options.allowClone,
                cloneCallback: cloneCallback,
                allowReorder: this.options.allowReorder,
                deleteConfirmation: this.options.deleteConfirmation,
                cloneConfirmation: this.options.cloneConfirmation
            });

            this.removeNoEntriesMsg();

            view.bind(this._event.ROW_EDIT, function (field) {
                AJS.triggerEvtForInst(instance._event.EDIT_ROW, instance, [this]);
                AJS.triggerEvtForInst(this._event.EDIT_ROW, {}, [this, instance]);
                instance.edit(this, field);
            });

            $row = view.render().$el;

            if (index !== -1) {

                if (typeof index === "number" && $rows.length !== 0 && index < $rows.length) {
                    $row.insertBefore($rows[index]);
                } else {
                    this.$tbody.append($row);
                    view.trigger(instance._event.ROW_APPENDED_TO_DOM);
                }
            }

            $row.data(this.dataKeys.ROW_VIEW, view);

            // deactivate all rows - used in the cases, such as opening a dropdown where you do not want the table editable
            // or any interactions
            view.bind(this._event.MODAL, function () {
                instance.$table.removeClass(instance.classNames.ALLOW_HOVER);
                instance.$tbody.sortable("disable");
                $.each(instance.getRows(), function () {
                    if (!instance.isRowBeingEdited(this)) {
                        this.delegateEvents({}); // clear all events
                    }
                });
            });

            view.bind(this._event.ANIMATION_STARTED, function () {
                instance.$table.removeClass(instance.classNames.ALLOW_HOVER);
            });

            view.bind(instance._event.ROW_MOVED, function (row, index) {
                instance.moveRow(row, index);
                AJS.triggerEvtForInst(instance._event.REORDER_SUCCESS, instance, [instance]);
            });

            view.bind(this._event.ANIMATION_FINISHED, function () {
                instance.$table.addClass(instance.classNames.ALLOW_HOVER);
            });

            // activate all rows - used in the cases, such as opening a dropdown where you do not want the table editable
            // or any interactions
            view.bind(this._event.MODELESS, function () {
                instance.$table.addClass(instance.classNames.ALLOW_HOVER);
                instance.$tbody.sortable("enable");
                $.each(instance.getRows(), function () {
                    if (!instance.isRowBeingEdited(this)) {
                        this.delegateEvents(); // rebind all events
                    }
                });
            });

            // ensure that when this row is focused no other are
            this._applyFocusCoordinator(view);

            this.trigger(this._event.ROW_INITIALIZED, view);

            return view;
        },

        /**
         * Returns if the row is edit mode or note
         *
         * @param {RAVEN.RestfulTable.Row} - read onyl row to check if being edited
         * @return {Boolean}
         */
        isRowBeingEdited: function (row) {

            var isBeingEdited = false;

            $.each(this.editRows, function () {
                if (this.el === row.el) {
                    isBeingEdited = true;
                    return false;
                }
            });

            return isBeingEdited;
        },

        getRowBeingEdited: function (row) {
            var rowEdit = null;

            $.each(this.editRows, function () {
                if (this.el === row.el) {
                    rowEdit = this;
                    return false;
                }
            });

            return rowEdit;
        },


        /**
         * Ensures that when supplied view is focused no others are
         *
         * @param {Backbone.View} view
         * @return {RAVEN.RestfulTable}
         */
        _applyFocusCoordinator: function (view) {

            var instance = this;

            if (!view.hasFocusBound) {

                view.hasFocusBound = true;

                view.bind(this._event.FOCUS, function () {
                    if (instance.focusedRow && instance.focusedRow !== view) {
                        instance.focusedRow.trigger(instance._event.BLUR);
                    }
                    instance.focusedRow = view;
                    if (view instanceof RAVEN.RestfulTable.Row && instance._createRow) {
                        instance._createRow.enable();
                    }
                });
            }

            return this;
        },

        /**
         * Remove specificed row from collection holding rows being concurrently edited
         *
         * @param {RAVEN.RestfulTable.EditRow} editView
         * @return {RAVEN.RestfulTable}
         */
        _removeEditRow: function (editView) {
            var index = $.inArray(editView, this.editRows);
            this.editRows.splice(index, 1);
            return this;
        },

        /**
         * Focuses last row still being edited or create row (if it exists)
         *
         * @return {RAVEN.RestfulTable}
         */
        _shiftFocusAfterEdit: function () {

            if (this.editRows.length > 0) {
                this.editRows[this.editRows.length-1].trigger(this._event.FOCUS);
            } else if (this._createRow) {
                this._createRow.trigger(this._event.FOCUS);
            }

            return this;
        },

        /**
         * Evaluate if we save row when we blur. We can only do this when there is one row being edited at a time, otherwise
         * it causes an infinate loop JRADEV-5325
         *
         * @return {boolean}
         */
        _saveEditRowOnBlur: function () {
            return this.editRows.length <= 1;
        },

        /**
         * Dismisses rows being edited concurrently that have no changes
         */
        dismissEditRows: function () {
            var instance = this;
            $.each(this.editRows, function () {
                if (!this.hasUpdates()) {
                    this.trigger(instance._event.FINISHED_EDITING);
                }
            });
        },

        /**
         * Converts readonly row to editable view
         *
         * @param {Backbone.View} row
         * @param {String} field - field name to focus
         * @return {Backbone.View} editRow
         */
        edit: function (row, field) {
        	 
            var instance = this,
                    editRow = new this.options.views.editRow({
                        el: row.el,
                        columns: this.options.columns,
                        isUpdateMode: true,
                        allowReorder: this.options.allowReorder,
                        model: row.model,
                        cancelAccessKey: this.options.cancelAccessKey,
                        submitAccessKey: this.options.submitAccessKey
                    }),
                    values = row.model.toJSON();
            values.index = row.el.attributes['data-index'].nodeValue;
            row.model.attributes.index = values.index;
            values.update = true;
            editRow.render({
                errors: {},
                update: true,
                values: values
            })
                    .bind(instance._event.UPDATED, function (model, focusUpdated) {
                        instance._removeEditRow (this);
                        this.unbind();
                        row.render().delegateEvents(); // render and rebind events
                        row.trigger(instance._event.UPDATED); // trigger blur fade out
                        if (focusUpdated !== false) {
                            instance._shiftFocusAfterEdit();
                            if (instance.sameOrder) {
                            	AJS.$("."+this.classNames.DRAG_HANDLE).show();
                            }
                        }else {
                        	if (instance.sameOrder) {
                        		AJS.$("."+this.classNames.DRAG_HANDLE).hide();
                        	}
                        }
                        AJS.triggerEvtForInst(instance._event.ROW_UPDATED, instance, [this]);
                    })
                    .bind(instance._event.VALIDATION_ERROR, function () {
                        this.trigger(instance._event.FOCUS);
                        if (instance.sameOrder) {
                        	AJS.$("."+this.classNames.DRAG_HANDLE).show();
                        }
                    })
                    .bind(instance._event.FINISHED_EDITING, function () {
                        instance._removeEditRow(this);
                        row.render().delegateEvents();
                        this.unbind();  // avoid any other updating, blurring, finished editing, cancel events being fired
                        if (instance.sameOrder) {
                        	AJS.$("."+this.classNames.DRAG_HANDLE).show();
                        }
                    })
                    .bind(instance._event.CANCEL, function () {
                        instance._removeEditRow(this);
                        this.unbind();  // avoid any other updating, blurring, finished editing, cancel events being fired
                        row.render().delegateEvents(); // render and rebind events
                        instance._shiftFocusAfterEdit();
                        
                        if (instance.sameOrder) {
                        	AJS.$("."+this.classNames.DRAG_HANDLE).show();
                        }
                    })
                    .bind(instance._event.BLUR, function () {
                        instance.dismissEditRows(); // dismiss edit rows that have no changes
                        if (instance._saveEditRowOnBlur()) {
                            this.trigger(instance._event.SAVE, false);  // save row, which if successful will call the updated event above
                        }
                    });
            // Ensure that if focus is pulled to another row, we blur the edit row
            this._applyFocusCoordinator(editRow);

            // focus edit row, which has the flow on effect of blurring current focused row
            editRow.trigger(instance._event.FOCUS, field);

            // disables form fields
            if (instance._createRow) {
                instance._createRow.disable();
            }

            this.editRows.push(editRow);
            
            return editRow;
        },


        /**
         * Renders all specified rows
         *
         * @param {Array} array of objects describing Backbone.Model's to render
         * @return {RAVEN.RestfulTable}
         */
        renderRows: function (rows) {
            var comparator = this._models.comparator, els = [], rowViews = [];

            this._models.comparator = undefined; // disable temporarily, assume rows are sorted

            var models = _.map(rows, function(row) {
                var model = new this.options.model(row);
                var view = this._renderRow(model, -1);
                rowViews.push(view);
                els.push(view.el);
                return model;
            }, this);
            this._models.add(models, {silent:true});

            this._models.comparator = comparator;

            this.removeNoEntriesMsg();

            this.$tbody.append(els);

            return rowViews;
        },

        /**
         * Gets default options
         *
         * @param {Object} options
         */
        _getDefaultOptions: function (options) {
            return {
                model: options.model || RAVEN.RestfulTable.EntryModel,
                allowEdit: true,
                allowClone: false,
                views: {
                    editRow: RAVEN.RestfulTable.EditRow,
                    row: RAVEN.RestfulTable.Row
                },
                Collection: Backbone.Collection.extend({
                    url: options.resources.self,
                    model: options.model || RAVEN.RestfulTable.EntryModel
                }),
                allowReorder: false,
                loadingMsg: options.loadingMsg || AJS.I18n.getText("aui.words.loading")
            }
        },

        _reindexRows: function() {
            var instance = this;

    		$.each(instance.getRows(), function (i, row) {
                var index = i + 1; 

                var editRow = instance.getRowBeingEdited(row);
                if (editRow) {
                    editRow.model.set("index", index);
                    editRow.$el.children('td:nth-child(2)').find('div.index').text(index);
                }
                row.model.set("index", index);
                row.$el.children('td:nth-child(2)').find('div.index').text(index);
                row.$el.attr('data-index', index);
            });
    	}
    });

    // jQuery data keys (http://api.jquery.com/jQuery.data/)
    RAVEN.RestfulTable.DataKeys = {
        ENABLED_SUBMIT: "enabledSubmit",
        ROW_VIEW: "RestfulTable_Row_View"
    };

    // CSS style classes. DON'T hard code
    RAVEN.RestfulTable.ClassNames = {
        NO_VALUE: "raven-restfultable-editable-no-value",
        NO_ENTRIES: "raven-restfultable-no-entires",
        RESTFUL_TABLE: "raven-restfultable",
        ROW: "raven-restfultable-row",
        READ_ONLY: "raven-restfultable-readonly",
        ACTIVE: "raven-restfultable-active",
        ALLOW_HOVER: "raven-restfultable-allowhover",
        FOCUSED: "raven-restfultable-focused",
        MOVEABLE: "raven-restfultable-movable",
        ANIMATING: "raven-restfultable-animate",
        DISABLED: "raven-restfultable-disabled",
        SUBMIT: "raven-restfultable-submit",
        EDIT_ROW: "raven-restfultable-editrow",
        CREATE: "raven-restfultable-create",
        DRAG_HANDLE: "raven-restfultable-draghandle",
        ORDER: "raven-restfultable-order",
        CONTENT: "raven-restfultable-content",
        EDITABLE: "raven-restfultable-editable",
        FOCUSABLE: "raven-restfultable-focusable",
        ERROR: "error",
        DELETE: "raven-restfultable-delete",
        CLONE: "raven-restfultable-clone",
        LOADING: "loading",
        TOOLBAR: "raven-restfultable-toolbar"
    };

    // Custom events
    RAVEN.RestfulTable.Events = {

        // AJS events
        REORDER_SUCCESS: "RestfulTable.reorderSuccess",
        ROW_ADDED: "RestfulTable.rowAdded",
        ROW_REMOVED: "RestfulTable.rowRemoved",
        ROW_UPDATED: "RestfulTable.rowUpdated",
        ROW_MOVED: "RestfulTable.rowMoved",
        EDIT_ROW: "RestfulTable.switchedToEditMode",
        SERVER_ERROR: "RestfulTable.serverError",

        // backbone events
        CREATED: "created",
        UPDATED: "updated",
        FOCUS: "focus",
        BLUR: "blur",
        SUBMIT: "submit",
        SAVE: "save",
        MODAL: "modal",
        MODELESS: "modeless",
        CANCEL: "cancel",
        CONTENT_REFRESHED: "contentRefreshed",
        RENDER: "render",
        FINISHED_EDITING: "finishedEditing",
        VALIDATION_ERROR: "validationError",
        SUBMIT_STARTED: "submitStarted",
        SUBMIT_FINISHED: "submitFinished",
        ANIMATION_STARTED: "animationStarted",
        ANIMATION_FINISHED: "animationFinisehd",
        INITIALIZED: "initialized",
        ROW_INITIALIZED: "rowInitialized",
        ROW_EDIT: "editRow",
        ROW_APPENDED_TO_DOM: "rowAppendedToDom",
        EDIT_FOCUS: 'editFocus'
    };

})(AJS.$);